// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

var EXPORTED_SYMBOLS = ["onlineFavoritesBackend"];

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;
const NAME = "http://home.netscape.com/NC-rdf#Name";

CU.import("resource:///modules/FlockScheduler.jsm");

/**************************************************************************
 * Module: Online Favorites Back End
 **************************************************************************/

/* Note: This module depends on access to the following PRIVATE properties
         of the services being supported:
         - aService._coop
         - aService._logger
         - aService._c_svc
*/

// Bookmark stream refresh interval in seconds.
const DEFAULT_REFRESH_INTERVAL = 5 * 60;
const PREF_ONLINEFAVE_REFRESH_INTERVAL = "flock.favorites.online.refreshInterval";

CC["@mozilla.org/moz/jssubscript-loader;1"]
.getService(CI.mozIJSSubScriptLoader)
.loadSubScript("chrome://flock/content/common/flocksafe.js");


/**************************************************************************
 * Private Data and Functions
 **************************************************************************/

// Helper function to assist in comparing strings.
function stringify(aObj) {
  if (!aObj) {
    return "";
  } else {
    return aObj;
  }
}

// Helper function to assist in comparing dates.
function dateify(aObj) {
  if (!aObj) {
    return 0;
  } else {
    return aObj.getTime();
  }
}

// Helper function to assist in comparing lists of tags.
function sanitizeTags(aTagList) {
  return aTagList ? aTagList.split(/[\s,]/).sort().join(",") : "";
}

/**************************************************************************
 * Online Favorites Back End Public Interface
 **************************************************************************/

var onlineFavoritesBackend = {};
var timersHash = {};

onlineFavoritesBackend.createAccount =
function OFBE_createAccount(aService, aAccountID, aIsTransient) {
  var account_urn = aService.urn + ":" + aAccountID;
  var timers = this.getTimers(account_urn);
  if (timers.length > 0) {
    // the user just clicked "Forget Account" before. 
    // so abort the deletion of bookmarks
    FlockScheduler.cancel(timers, 0);
  }
  
  if (!aService._coop.Account.exists(account_urn)) {
    var account = new aService._coop.Account(account_urn, {
      name: aAccountID,
      serviceId: aService.contractId,
      service: aService._c_svc,
      accountId: aAccountID,
      favicon: aService.icon,
      URL: aService.getUserUrl(aAccountID),
      isTransient: aIsTransient,
      showInSidebar: false
    });
    aService._coop.accounts_root.children.addOnce(account);
  }

  // Create a "stream" for the online bookmarks
  var bookmarks_urn = account_urn + ":bookmarks";
  var bookmarks = aService._coop.get(bookmarks_urn);
  if (!bookmarks) {
    prefService = CC["@mozilla.org/preferences-service;1"]
                  .getService(CI.nsIPrefBranch);
    var prefRefreshInterval = DEFAULT_REFRESH_INTERVAL;
    if (prefService.getPrefType(PREF_ONLINEFAVE_REFRESH_INTERVAL)) {
      prefRefreshInterval = prefService.getIntPref(PREF_ONLINEFAVE_REFRESH_INTERVAL);
    }

    bookmarks = new aService._coop.OnlineBookmarksStream(bookmarks_urn, {
      name: aAccountID + " on " + aService.title,
      userid: aAccountID,
      serviceId: aService.contractId,
      favicon: aService.icon,
      isTransient: aIsTransient,
      refreshInterval: prefRefreshInterval
    });
  } else {
    bookmarks.isPollable = true;
    bookmarks.nextRefresh = Date.now();
    bookmarks.last_update_time = new Date(0);
  }
  //make sure it's added to its parents (old profiles can be corrupted due to old sync problem)
  aService._coop.onlinebookmarks_root.children.addOnce(bookmarks);

  var noTagFolder_urn = bookmarks.id() + ":notag";
  var noTagFolder = aService._coop.get(noTagFolder_urn);
  if (!noTagFolder) {
    // Create the "no tag" folder
    noTagFolder = new aService._coop.Folder(noTagFolder_urn, {
      name: flockGetString("favorites/favorites", "flock.favs.online.noTag")
    });
  }
  if (!bookmarks.children.has(noTagFolder)) {
    bookmarks.children.insertAt(noTagFolder, 1);
  }

  var acct = aService.getAccount(account_urn);
  return acct;
};

onlineFavoritesBackend.removeAccount =
function OFBE_removeAccount(aService, aAccountUrn) {
  var timers = this.getTimers(aAccountUrn);
  if (timers.length > 0) {
    FlockScheduler.cancel(timers, 0);
    aService._logger.debug("Aborted update of " + aAccountUrn);
  }

  var acctCoopObj = aService._coop.get(aAccountUrn);
  // Let's kill the account in order to give feedback right away to 
  // the user. Coop will also remove it from its parent(s).
  acctCoopObj.destroy();

  var bookmarks_urn = aAccountUrn + ":bookmarks";
  var bookmarks = aService._coop.get(bookmarks_urn);

  // Slowly delete the bookmarks that are associated with this account
  var synchronize = function OFBE_removeAccount_sync(should_yield) {
    var bmChildren = bookmarks.children.enumerate();
    while (bmChildren.hasMoreElements()) {
      var tag = bmChildren.getNext();
      var enum_ = tag.children.enumerate();
      while (enum_.hasMoreElements()) {
        var bm = enum_.getNext();
        aService._logger.debug("destroying bookmark " + bm.URL);
        bm.destroy();
        if (should_yield()) {
          yield;
        }
      }
      tag.destroy();
    }
    bookmarks.destroy();
    aService._logger.debug("Destroyed account " + aAccountUrn);
  };

  bookmarks.isPollable = null; // stop polling for updates
  FlockScheduler.schedule(timers, 0.1, 10, synchronize);
};

onlineFavoritesBackend.updateBookmark =
function OFBE_updateBookmark(aService, aAccountUrn, aServerBookmark, 
                             aLocalBookmarks, aPrivate) {
  var bookmark_urn = aAccountUrn + ":" + aServerBookmark.URL;
  var tagList;
  var i;

  var localBookmark = aService._coop.get(bookmark_urn);
  // Create/update the bookmark itself
  if (localBookmark) {
    // Need to check if any modification is needed before we assert anything
    var modified = false;
    if (stringify(aServerBookmark.name) !== stringify(localBookmark.name)) {
      modified = true;
      localBookmark.name = aServerBookmark.name;
    }
    if (stringify(aServerBookmark.description) !==
        stringify(localBookmark.description))
    {
      modified = true;
      localBookmark.description = aServerBookmark.description;
    }
    var sanitizedServerTags = sanitizeTags(aServerBookmark.tags);
    if (sanitizedServerTags != sanitizeTags(localBookmark.tags))
    {
      modified = true;
      localBookmark.tags = sanitizedServerTags;
      localBookmark.tag.removeAll();
    }
    if (localBookmark.private != aPrivate) {
      modified = true;
      localBookmark.private = aPrivate;
    }
    if (dateify(aServerBookmark.datevalue) == dateify(localBookmark.datevalue)
        && !modified) {
      //check to fix old corrupted profiles.
      var firstTag = localBookmark.getParent();
      if (firstTag && firstTag.getParent()) {
        aService._logger.debug("Unchanged on server: Ignoring "
                               + aServerBookmark.URL);
        return;
      }
    }
    aService._logger.debug("Modified on server : Updating "
                           + aServerBookmark.URL);
    var date = new Date();
    localBookmark.LastModifiedDate = date;
  } else {
    // New bookmark
    aService._logger.debug("New on server      : Creating "
                           + aServerBookmark.URL);
    localBookmark = new aService._coop.Bookmark(bookmark_urn, {
      URL: aServerBookmark.URL,
      name: aServerBookmark.name,
      description: aServerBookmark.description,
      tags: aServerBookmark.tags,
      datevalue: aServerBookmark.datevalue,
      hash: aServerBookmark.hash,
      flockType: "bookmark",
      private: aPrivate,
      isPollable: false
    });
    // bookmarks.children.add(localBookmark);
  }
  tagList = aServerBookmark.tags.split(/[\s,]/);
  for (i = 0; i < tagList.length; i++) {
    localBookmark.tag.addOnce(tagList[i]);
  }
  if (!tagList.length) {
    tagList = [""];
  }
  this.classifyBookmark(aService, aLocalBookmarks, localBookmark, tagList);
};

onlineFavoritesBackend.classifyBookmark =
function OFBE_classifyBookmark(aService, aLocalBookmarks,
                               aLocalBookmark, aTagList) {
  var noTagFolder = aService._coop.get(aLocalBookmarks.id() + ":notag");
  // Delete the bookmark from the tag folders, and delete empty tags
  var tags = aLocalBookmark.getParents();
  while (tags.length > 0) {
    var tag_ = tags.pop();
    var i = aTagList.indexOf(tag_.name);
    if (i >= 0 && tag_.getParent()) {
      // don't need to change this tag.
      aTagList.splice(i, 1);
    } else {
      tag_.children.remove(aLocalBookmark);
      //do not use children.count() because it's not updated after remove()
      if (tag_ != noTagFolder && 
          !tag_.children.enumerate().hasMoreElements()) {
        tag_.destroy();
      }
    }
  }

  // Add the bookmark to the tag folders it should be
  for each (var tag in aTagList) {
    var tagUrn = aLocalBookmarks.id() + ":tag:" + tag;
    var tagFolder;
    var insertIt = true;
    if (tag == "") {
      tagFolder = noTagFolder;
      insertIt = false;
    } else {
      tagFolder = aService._coop.get(tagUrn, true);
      if (!tagFolder) {
        tagFolder = new aService._coop.Folder(tagUrn, { name: tag });
      } else {
        //check to fix corrupted profiles.
        insertIt = !aLocalBookmarks.children.has(tagFolder);
      } // else also insert it to fix corrupted profiles.
    }
    tagFolder.children.addOnce(aLocalBookmark);
    if (insertIt) {
      //aService._logger.debug("inserting Tag " + tag);
      aLocalBookmarks.children.insertSortedOn(NAME, tagFolder);
      if (aLocalBookmarks.children.indexOf(noTagFolder) != 1) {
        aLocalBookmarks.children.remove(noTagFolder);
        aLocalBookmarks.children.insertAt(noTagFolder, 1);
      }
    }
  }
};

onlineFavoritesBackend.destroyBookmark =
function OFBE_destroyBookmark(aService, aAccountId, aUrl) {
  var bookmark_urn = aService.urn + ":" + aAccountId + ":" + aUrl;

  if (aService._coop.Bookmark.exists(bookmark_urn)) {
    var bookmarks = aService._coop.get(aService.urn + ":"
                  + aAccountId + ":bookmarks");
    var bookmark = aService._coop.get(bookmark_urn);

    bookmark.destroy();
  }
};

onlineFavoritesBackend.getTimers =
function OFBE_getTimers(aAccountUrn) {
  var timers = timersHash[aAccountUrn];
  if (!timers) {
    timers = [];
    timersHash[aAccountUrn] = timers;
  }
  return timers;
};

onlineFavoritesBackend.updateLocal =
function OFBE_updateLocal(aService, aPostsList, aLastUpdate, aAccountUrn) {
  var timers = this.getTimers(aAccountUrn);
  if (timers.length > 0) {
    // let the previous update finish.
    aService._logger.debug("Ignoring refresh of " + aAccountUrn);
    return;
  }
  var localBookmarks = aService._coop.get(aAccountUrn + ":bookmarks");
  if (!localBookmarks || !localBookmarks.isPollable) {
    return; // removeAccount() has just been called
  }

  var i;
  var bmEnum;
  var bmEnum2;

  var be = this;
  var synchronize = function OFBE_updateLocal_sync(should_yield) {
    var urlHash = {}; // Useful to optimize deletion

    // Create new bookmarks, or update existing bookmarks
    for (i = 0; i < aPostsList.length; i++) {
      if (aPostsList[i].tags === "system:unfiled") {
        aPostsList[i].tags = "";
      }
      // To optimize deletion: put the URLs in a hash
      urlHash[aPostsList[i].URL] = true;
      be.updateBookmark(aService, aAccountUrn, aPostsList[i], localBookmarks);
      if (should_yield()) {
        yield;
        var trigger = new Date(Date.now() + DEFAULT_REFRESH_INTERVAL * 1000);
        if (trigger > localBookmarks.nextRefresh ) {
          // make sure that we have time to finish this update 
          // before another one comes
          localBookmarks.nextRefresh = new Date(trigger.getTime() + 
                                       localBookmarks.refreshInterval * 1000);
          aService._logger.debug("Pushing next refresh to " + 
                                 localBookmarks.nextRefresh);
        }
      }
    }

    // Remove locally bookmarks deleted on the server
    var localLength = 0;
    i = 0;
    bmEnum = localBookmarks.children.enumerate();
    while (bmEnum.hasMoreElements()) {
      var localTag = bmEnum.getNext();
      bmEnum2 = localTag.children.enumerate();
      var tagIsEmpty = true;
      while (bmEnum2.hasMoreElements()) {
        var localBookmark = bmEnum2.getNext();
        if (!urlHash[localBookmark.URL]) {
          aService._logger.debug("Deleted on server  : Deleting "
                                 + localBookmark.URL);
          localBookmark.destroy();
        } else {
          tagIsEmpty = false;
        }
      }
      // Remove empty tags
      if (tagIsEmpty && localTag.id() != localBookmarks.id() + ":notag") {
        // The tag is empty, let's delete it.
        // Note: Don't delete "No Tag".
        localTag.destroy();
      }
      if (should_yield()) {
        yield;
      }
    }

    // Update the last update time for the bookmark stream.
    // Do it last in case the user shuts down before it's finished.
    localBookmarks.last_update_time = aLastUpdate;
  };

  FlockScheduler.schedule(timers, 0.1, 10, synchronize);
};

onlineFavoritesBackend.updateTags =
function OFBE_updateTags(aService, aAccountId, aTags) {
  var localBookmarks = aService._coop.get(aService.urn + ":"
                     + aAccountId + ":bookmarks");
  if (!localBookmarks || !localBookmarks.isPollable) {
    return; // destroy is in progress
  }
  // Copy aTags, so we can modify it.
  var tags = [];
  var tag;
  var i;
  for (i = 0; i < aTags.length; i++) {
    tags.push(aTags[i]);
  }
  // Remove tags no longer present.
  var tagenum = localBookmarks.tag.enumerate();
  while (tagenum.hasMoreElements()) {
    tag = tagenum.getNext();
    if (tags.indexOf(tag) < 0) {
      // Not in new array, so remove it.
      localBookmarks.tag.remove(tag);
    } else {
      // Already there, remove from input so we don't try adding it again.
      tags.splice(tags.indexOf(tag), 1);
    }
  }
  // Put in the net new tags.
  for (i = 0; i < tags.length; i++) {
    localBookmarks.tag.addOnce(tags[i]);
  }
};

onlineFavoritesBackend.showNotification =
function OFBE_showNotification(aMessage) {
  var wm = CC["@mozilla.org/appshell/window-mediator;1"]
           .getService(CI.nsIWindowMediator);
  var topNavWindow = wm.getMostRecentWindow("navigator:browser");

  var nBox = topNavWindow.gBrowser.getNotificationBox();
  var notification = nBox.getNotificationWithValue("favorite-error");
  if (notification) {
    notification.label = aMessage;
  } else {
    nBox.appendNotification(aMessage,
                            "favorite-error",
                            "chrome://browser/skin/Info.png",
                            nBox.FLOCK_PRIORITY_HIGH,
                            null);
  }
};
